"""
Copyright 2012 NetApp, Inc. All Rights Reserved,
contribution by Weston Andros Adamson <dros@netapp.com>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
"""

import os, sys, time
import re
import numpy as np

from config import *
from selector import Selector
import parse
from trace import TraceAttrs

class StatBin:
    """ a collection of statistics pulled from a file """
    def __init__(self, name, description, workload, filename, collection):
        self.name = name
        self.description = description
        self.workload = workload
        self.filename = filename
        self.collection = collection
        self._values = {}
        self._workloaddirs = [] # MUST be same # of entries per value

        self._changed = False

    def keys(self):
        """ return all keys contained in this object """
        return self._values.keys()

    def has_key(self, k):
        return self._values.has_key(k)

    def add_workloaddir(self, workloaddir):
        """ add a workload result directory.  this MUST be done for each set of 
            values """
        self._changed = True
        self._workloaddirs.append(workloaddir)

    def add(self, key, value, units, descr, better):
        """ add a value for the key.  should be called once on each key for
            every workload result directory """
        self._changed = True

        if not self._values.has_key(key):
            self._values[key] = []

        self._values[key].append(value)
        self.collection.set_valid_statbin_key(self.name, key)
        self.collection.set_stat_info(self.name, key, units, descr, better)

    def balance(self, source=None):
        """ makes sure that the lists are all the same length """

        # mountstats(1) doesnt show ops with 0 counts
        def _suppress(key):
            for end in ['_count',
                        '_rtt',
                        '_exectime',
                        '_avg_bytes_sent',
                        '_avg_bytes_received']:
                if key.endswith(end):
                    return True
            return False

        if self._changed:
            expect = len(self._workloaddirs)
            for k, v in self._values.iteritems():

                if expect > len(v):
                    v.extend([0.0] * (expect - len(v)))
                    if not _suppress(k):
                        warn("zerofilled %s:%s:%s from %s" %
                                (self.workload, self.name, k, source))

            self._changed = False

    def _check(self):
        #assert not self._changed
        if self._changed:
            self.balance()

    def get(self, key, *args):
        """ get values for key, returns list of values """
        self._check()

        if len(args):
            assert len(args) == 1
            return self._values.get(key, args[0])
        else:
            return self._values[key]

    def num_runs(self):
        """ returns the number of runs contained in this object """
        self._check()
        keys = self._values.keys()

        if keys:
            return len(self._values[keys[0]])

        return 0

    def get_run(self, index, keys=None):
        """ returns the referenced run's values for all keys """
        self._check()

        assert index < self.num_runs()

        vals = {}

        if keys == None:
            for k, v in self._values.iteritems():
                vals[k] = v[index]
        else:
            for k in keys:
                if self._values.has_key(k):
                    vals[k] = self._values[k][index]

        return vals

class TraceStats:
    """ a collection of StatBin objects """
    def __init__(self, selection, resultsdir, collection):
        self.resultsdir = resultsdir

        self.selection = selection
        self.collection = collection

        self.statbins = {}
        self.attrs = {}
        self.workloaddirs = []

    def merge_attrs(self, new):
        need_attrs = ['client', 'server', 'path', ]

        if not self.attrs:
            self.attrs = dict([(k,new[k]) for k in need_attrs])

        for k in need_attrs:
            assert self.attrs.has_key(k)
            assert self.attrs[k] == new[k], \
                    "key %s mismatch: %r != %r" % (k, self.attrs[k], new[k])

    def ref_statbin(self, name, description, filename):
        if not self.statbins.has_key(name):
            self.statbins[name] = StatBin(name, description,
                                         self.selection.workload,
                                         filename,
                                         self.collection)
        return self.statbins[name]

    def get_statbin(self, name):
        return self.statbins[name]

    def get_statbins(self):
        return self.statbins.keys()

    def add_workloaddir(self, workloaddir):
        self.workloaddirs.append(workloaddir)

    def num_runs(self):
        return len(self.workloaddirs)

class TraceCollection:
    """ A collection of TraceStats objects, with interface to load them
        from `resultsdir' """

    def __init__(self, resultsdir):
        assert os.path.isdir(resultsdir)
        self.resultsdir = resultsdir

        self._tracestats = {}
        self._stat_units = {}
        self._stat_descriptions = {}
        self._stat_betters = {}
        self._valid_statbin_keys = {}
        self._used_statbin_keys = {}

        cwd = os.getcwd()
        os.chdir(self.resultsdir)

        for ent in os.listdir('.'):
            try:
                # old
                if ent.startswith('test-') and os.path.isdir(ent):
                    self.load_tracedir(ent)
                # also old
                elif ent.startswith('nfstest-') and os.path.isdir(ent):
                    self.load_tracedir(ent)
                # new
                elif ent.startswith(TRACE_DIR_PREFIX) and os.path.isdir(ent):
                    self.load_tracedir(ent)
            except:
                print "Error loading workloaddir: %s" % ent
                raise

        os.chdir(cwd)

        workloads = {}
        kernels = {}
        nfsopts = {}
        clients = {}
        servers = {}

        for sel in self._tracestats.keys():
            workloads[sel.workload] = True
            kernels[sel.kernel] = True
            nfsopts[sel.nfsopt] = True
            clients[sel.client] = True
            servers[sel.server] = True

        # get sorting out of the way now
        workloads = workloads.keys()
        workloads.sort()
        kernels = kernels.keys()
        kernels.sort()
        nfsopts = nfsopts.keys()
        nfsopts = nfsopt_normalize(nfsopts)
        clients = clients.keys()
        clients.sort()
        servers = servers.keys()
        servers.sort()

        self.selection = Selector(workloads, kernels, nfsopts, clients, servers)

    def is_empty(self):
        return len(self._tracestats) == 0

    def set_stat_info(self, statbin, key, units, descr, better):
        if not self._stat_units.has_key((statbin, key)):
            self._stat_units[(statbin, key)] = units
        else:
            assert self._stat_units[(statbin, key)] == units

        if not self._stat_descriptions.has_key((statbin, key)):
            self._stat_descriptions[(statbin, key)] = descr
        else:
            assert self._stat_descriptions[(statbin, key)] == descr

        if not self._stat_betters.has_key((statbin, key)):
            self._stat_betters[(statbin, key)] = better
        else:
            assert self._stat_betters[(statbin, key)] == better

    def stat_description(self, statbin, key, descmap={}):
        if descmap and descmap.has_key(key):
            return descmap[key]
        return self._stat_descriptions.get((statbin, key), None)

    def stat_units(self, statbin, key):
        return self._stat_units.get((statbin, key), None)

    def stat_better(self, statbin, key, bettermap={}):
        if bettermap:
            better = bettermap.get(key, BETTER_UNKNOWN)
        else:
            better = self._stat_betters.get((statbin, key), BETTER_UNKNOWN)

        return better

    def get_better_info(self, selection, statbin, key, bettermap={}):
        bounds = TEST_BOUND_IO
        # XXX should come from workload definition
        if selection.workload.startswith('filebench_'):
            bounds = TEST_BOUND_TIME

        better = self.stat_better(statbin, key, bettermap=bettermap)

        return better_info(bounds, better)

    def set_valid_statbin_key(self, statbin, key):
        if not self._valid_statbin_keys.has_key(statbin):
            self._valid_statbin_keys[statbin] = set()
        self._valid_statbin_keys[statbin].add(key)

    def get_valid_statbin_keys(self, statbin):
        keys = list(self._valid_statbin_keys.get(statbin, set()))
        keys.sort()
        return keys

    def set_used_statbin_keys(self, statbin, keys):
        if not self._used_statbin_keys.has_key(statbin):
            self._used_statbin_keys[statbin] = set()
        for k in keys:
            self._used_statbin_keys[statbin].add(k)

    def get_used_statbin_keys(self, statbin):
        keys = list(self._used_statbin_keys.get(statbin, set()))
        keys.sort()
        return keys

    def load_info(self):
        nrun_map = {}
        total_runs = 0

        for sel in self.selection.foreach():
            try:
                tracestat = self.get_trace(sel)
                nruns = tracestat.num_runs()
            except KeyError:
                nruns = 0

            total_runs += nruns

            if not nrun_map.has_key(nruns):
                nrun_map[nruns] = []
            nrun_map[nruns].append(sel)

        if nrun_map.keys():
            maxruns = max(nrun_map.keys())
            minruns = min(nrun_map.keys())
        else:
            maxruns = 0
            minruns = 0

        if maxruns == minruns:
            runstr = '%u run%s' % (maxruns, pluralize(maxruns))
        else:
            runstr = '%u-%u runs' % (minruns, maxruns)

        lk = len(self.selection.kernels)
        ln = len(self.selection.nfsopts)
        lw = len(self.selection.workloads)

        info = {
            'resultdir' : self.resultsdir,
            'run_str'   : runstr,
            'total'     : total_runs,
            'total_str' : 'trace%s' % (pluralize(total_runs),),
            'kernel'    : lk,
            'kernel_str': 'kernel%s' % pluralize(lk),
            'nfsopt'    : ln,
            'nfsopt_str': 'nfsopt%s' % pluralize(ln),
            'workload'  : lw,
            'workload_str' : 'workload%s' % pluralize(lw),
        }

        info['summary'] = '%s %s X %s %s X %s %s X %s = %s %s' % \
            ( info['kernel'], info['kernel_str'],
             info['nfsopt'], info['nfsopt_str'],
             info['workload'], info['workload_str'],
             info['run_str'],
             info['total'], info['total_str'])

        info['warnings'] = []
        for i in range(maxruns - 1, minruns - 1, -1):
            for s in nrun_map.get(i, []):
                info['warnings'].append('%s has %u run%s' % (s, i, pluralize(i)))

        return info

    def load_info_str(self):
        o = []
        info = self.load_info()

        o.append(info['summary'])
        o.append('')

        o.append('%s:' % info['kernel_str'])
        for k in self.selection.kernels:
            o.append("  %s" % k)

        o.append("%s:" % info['nfsopt_str'])
        for n in self.selection.nfsopts:
            o.append("  %s" % n)

        o.append("%s:" % info['workload_str'])
        for w in self.selection.workloads:
            o.append("  %s" % w)

        o.append('')
        for w in info['warnings']:
            o.append('  %s' % w)

        return '\n'.join(o)

    def _ref_trace(self, workload, kernel, nfsopts, client, server):
        """ return instance to TraceStats keyed by arguments """
        nfsopts = nfsopt_normalize(nfsopts)
        sel = Selector(workload, kernel, nfsopts, client, server)

        assert sel.is_valid_key(), "Invalid key: %r" % sel

        if not self._tracestats.has_key(sel):
            self._tracestats[sel] = TraceStats(sel, self.resultsdir, self)

        return self._tracestats[sel]

    def get_trace(self, selection):
        return self._tracestats[selection]

    def has_trace(self, selection):
        return self._tracestats.has_key(selection)

    def has_traces(self, selection):
        """ return True if this collection has any traces matching 'selection',
            otherwise returns False """
        for x in selection.foreach():
            if self._tracestats.has_key(x):
                return True
        return False

    def load_traceattrs(self, workloaddir):
        """ load attrs from attr file """
        attr = {'workloaddir': workloaddir,
               }

        attr_file = os.path.join(workloaddir, 'arguments')
        trace_attrs = TraceAttrs(filename=attr_file).to_dict()

        for k, v in trace_attrs.iteritems():
            attr[k] = v

        return attr

    def _check_dmesg(self, workloaddir):
        def _check_lines(f):
            return '\n'.join([ x for x in file(f).readlines()
                if x.startswith('>') and x.lower().find('nfs') >= 0 ])

        diff = os.path.join(workloaddir, 'dmesg.diff')
        result = _check_lines(diff)
        if result:
            return 'dmesg.start and dmesg.end are different:\n%s' % (result,)
        return ''

    def load_tracedir(self, workloaddir):
        """ load a trace directory and all stats contained within """
        assert os.path.isdir(workloaddir)

        attrs = self.load_traceattrs(workloaddir)

        warning = self._check_dmesg(workloaddir)
        if warning:
            warn("%s: %s" % (workloaddir, warning))

        if attrs['tags']:
            tags = attrs['tags'].split(',')
        else:
            tags = []

        nfsopts = make_nfsopt(attrs['mountopts'], tags)

        tracestat = self._ref_trace(attrs['workload'], attrs['kernel'], nfsopts,
                                    attrs['client'], attrs['server'])

        tracestat.add_workloaddir(workloaddir)
        tracestat.merge_attrs(attrs)

        try:
            parse.parse_time(tracestat, workloaddir)
        except:
            warn("error parsing times: %s" % workloaddir)

        try:
            parse.parse_mountstats(tracestat, workloaddir, nfsopts)
        except:
            warn("error parsing mountstats: %s" % workloaddir)

        try:
            parse.parse_nfsiostat(tracestat, workloaddir, attrs['localpath'])
        except:
            warn("error parsing nfsiostats: %s" % workloaddir)

        try:
            parse.parse_nfsstats(tracestat, workloaddir)
        except:
            warn("error parsing nfsstats: %s" % workloaddir)

        try:
            parse.parse_proc_mountstats(tracestat, workloaddir)
        except:
            if int(attrs['orig_tracedir_version']) > 4:
                warn("error parsing proc mountstats: %s" % workloaddir)

        if tracestat.selection.workload.startswith('filebench_'):
            try:
                parse.parse_filebench(tracestat, workloaddir)
            except:
                warn("error parsing filebench: %s" % workloaddir)
 
    def attrs(self, selection):
        attrs = {}

        for subsel in selection.foreach():
            try:
                tracestat = self.get_trace(subsel)
            except KeyError:
                continue

            for k, v in tracestat.attrs.iteritems():
                if not attrs.has_key(k):
                    attrs[k] = []
                if not v in attrs[k]:
                    attrs[k].append(v)

        for k, v in attrs.iteritems():
            # formatters
            _fmt = None
            if k == 'mount_options':
                _fmt = lambda x : x.split(',')

            if _fmt:
                new_v = []
                for x in v:
                    new_v.append(_fmt(x))
                v = new_v

            # collapse from lists if len == 1
            if len(v) == 1:
                attrs[k] = v[0]

        return attrs

    def _make_list(self, value, default):
        if value == None:
            return default
        elif isinstance(value, (list, tuple)):
            return value
        else:
            return [value]

    def num_runs_list(self, workload=None, kernel=None, nfsopt=None,
                      client=None, server=None):
        workload = self._make_list(workload, self.selection.workloads)
        kernel = self._make_list(kernel, self.selection.kernels)
        nfsopt = self._make_list(nfsopt, self.selection.nfsopts)
        client = self._make_list(client, self.selection.clients)
        server = self._make_list(server, self.selection.servers)

        selector = Selector(workload, kernel, nfsopt, client, server)

        num_runs_list = []
        for sel in selector.foreach():
            try:
                tracestat = self.get_trace(sel)
            except KeyError:
                continue
            num_runs_list.append(tracestat.num_runs())

        return num_runs_list

    def total_runs(self, kernel=None, workload=None, nfsopts=None):
        return sum(self.num_runs_list(kernel, workload, nfsopts))

    def is_stat_key(self, statbin, key):
        return key in self._valid_statbin_keys.get(statbin, set())

    def used_stat_key(self, statbin, key):
        return key in self._used_statbin_keys.get(statbin, set())

    def show_contents(self, pre=''):
        out = []
        for ssel in self.selection.foreach(('client', 'server')):
            out.append("%s%s -> %s" % (pre, ssel.client, ssel.server))
            for ksel in ssel.foreach('kernel'):
                out.append("%s kernel %s" % \
                    (pre, get_kernel_display(ksel.kernel),))
                for wsel in ksel.foreach('workload'):
                    if not self.has_traces(wsel):
                        continue

                    out.append("%s  %s" % \
                               (pre, get_workload_display(wsel.workload)))
                    for nsel in wsel.foreach('nfsopt'):
                        try:
                            tracestat = self.get_trace(nsel)
                        except:
                            continue
                        nruns = tracestat.num_runs()
                        out.append("%s   %u run%s of %s" % \
                            (pre,
                             nruns,
                             pluralize(nruns),
                             get_nfsopt_display(nsel.nfsopt)))
            out.append("%s" % (pre,))

        return out

