"""
Copyright 2012 NetApp, Inc. All Rights Reserved,
contribution by Weston Andros Adamson <dros@netapp.com>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
"""

import getopt
import re
import os, posix, stat, sys

from tags import *

CONFIG_MODULE_FILE=__file__

NFSOMETER_VERSION='1.3'

RUNNING_TRACE_DIR='/tmp/nfsometer_trace'
PROBE_DIR='/tmp/nfsometer_probe'

TRACE_ATTRFILE='arguments'
TRACE_DIR_PREFIX='nfsometer_trace'
TRACE_DIR_VERSION=6

TRACE_MOUNT_TRIES = 3
TRACE_MOUNT_TRY_DELAY = 1.0
TRACE_UMOUNT_TRIES = 60
TRACE_CLEANUP_UMOUNT_TRIES = 60
TRACE_UMOUNT_TRY_DELAY = 1.0

TRACE_LOADGEN_STAGGER_MAX = 60

NFSOMETERLIB_DIR=os.path.split(__file__)[0]
NFSOMETER_DIR=os.path.join(posix.environ['HOME'], '.nfsometer')
HTML_DIR="%s/html" % NFSOMETERLIB_DIR

# locations
MOUNTDIR=os.path.join(RUNNING_TRACE_DIR, 'mnt')
WORKLOADFILES_ROOT=os.path.join(NFSOMETER_DIR, 'workload_files')
RESULTS_DIR=os.path.join(posix.environ['HOME'], 'nfsometer_results')
HOSTNAME=posix.environ['HOSTNAME']
RUNROOT='%s/nfsometer_runroot_%s' % (MOUNTDIR, HOSTNAME)

TEMPLATE_TOC='%s/toc.html' % HTML_DIR
TEMPLATE_TOCNODE='%s/tocnode.html' % HTML_DIR
TEMPLATE_TABLE='%s/table.html' % HTML_DIR
TEMPLATE_DATASET='%s/dataset.html' % HTML_DIR
TEMPLATE_DATASETKEYINFO='%s/dataset_key_info.html' % HTML_DIR
TEMPLATE_WIDGET='%s/widget.html' % HTML_DIR
TEMPLATE_SECTION='%s/section.html' % HTML_DIR
TEMPLATE_REPORT='%s/report.html' % HTML_DIR
TEMPLATE_INDEX='%s/index.html' % HTML_DIR
TEMPLATE_REPORTLIST='%s/reportlist.html' % HTML_DIR
TEMPLATE_CELLDATAINFO='%s/cell_data_info.html' % HTML_DIR
TEMPLATE_LOADINFO='%s/load_info.html' % HTML_DIR

CSSFILEPATH='%s/style.css' % HTML_DIR
JSFILEPATH='%s/script.js' % HTML_DIR
JQUERY_URL='http://code.jquery.com/jquery-1.7.2.min.js'

GRAPH_ERRORBAR_COLOR='#CF2727'
GRAPH_ERRORBAR_WIDTH=1
GRAPH_EDGE_COLOR='#000000'

FILE_COMMAND='command.sh'

COMPARE_KERNELS='Kernel'
COMPARE_SERVERS='Server'

COLOR_TERMS = ('xterm-color', 'screen')

def highlight(msg):
    if posix.environ['TERM'] in COLOR_TERMS:
        pre = '\033[33m'
        post = '\033[0m'
    else:
        pre = '>> '
        post = ''

    print "%s%s%s" % (pre, msg, post)
    sys.stdout.flush()

def inform(msg):
    if posix.environ['TERM'] in COLOR_TERMS:
        pre = '\033[32m'
        post = '\033[0m'
    else:
        pre = '>> '
        post = ''

    print "%s%s%s" % (pre, msg, post)
    sys.stdout.flush()

def warn(msg):
    if posix.environ['TERM'] in COLOR_TERMS:
        pre = '\033[41m'
        post = '\033[0m'
    else:
        pre = 'WARNING> '
        post = ''

    print "%s%s%s" % (pre, msg, post)
    sys.stdout.flush()

def import_error(m):
     print >>sys.stderr, m
     sys.exit(1)


WORKLOADS_DIR=os.path.join(NFSOMETERLIB_DIR, 'workloads')
WORKLOADS_SCRIPT=os.path.join(WORKLOADS_DIR, 'workload.sh')

def get_workload_display(w, short=True):
    if isinstance(w, (list, tuple)):
        return ' '.join([ get_workload_display(x) for x in w ])
    return w

def get_nfsopt_display_long(nfsopt):
    return get_nfsopt_display(nfsopt, short=False)

def get_nfsopt_display(nfsopt, short=True):
    if isinstance(nfsopt, (list, tuple)):
        # special output if all versions are the same
        vers = None
        other = []
        for x in nfsopt:
            v, o = nfsopt_version_other(x)
            if vers and v != vers:
                vers = None
                other = []
                break
            vers = v
            other.append(o or v)

        if vers:
            if short:
                return "version %s" % (vers,)
            return "version %s - %s" % (vers, ' '.join(other))

        return ' '.join([ get_nfsopt_display(x, short=short)
                              for x in nfsopt ])

    # try to split off client args
    mt, st, ct = nfsopt_split(nfsopt)
    lookup = ','.join(mt)
    if st:
        lookup += ',' + ','.join(st)

    return nfsopt

def get_kernel_display(kernel, short=True):
    if isinstance(kernel, (list, tuple)):
        return ' '.join([ get_kernel_display(x) for x in kernel ])
    return kernel

#
# NOTE_* - disclaimers and such
#

# TODO this should be based off of some arg in workload def
#      for fixed-time tests
def note_filebench_times(sel):
    has_fb = False

    for w in sel.workloads:
        if w.startswith('filebench_'):
            has_fb = True
            break

    if has_fb:
        return """Filebench tests are run for a set amount
                  of time the <i>time_real</i> value is somewhat
                  useless.
               """
    return ''

def note_v3_no_lock(sel):
    for nfsopt in sel.nfsopts:
        if nfsopt_version(nfsopt) == 'v3':
            return """NFSv3's locking protocol runs on a different service
                      and is not counted
                   """
    return ''

def note_v41_pnfs_no_ds(sel):
    old_kernel = False
    for kernel in sel.kernels:
        # XXX a hack, and not really true since some versions < 3 do have these
        #     stats
        if kernel.startswith('2.'):
            old_kernel = True
            break

    if old_kernel:
        return """ Older linux kernels do not count READ, WRITE and
                   COMMIT operations to pNFS dataservers (unless the
                   DS is also the MDS)."""
    return ''
 
TEST_BOUND_UNKNOWN = 0
TEST_BOUND_IO = 1
TEST_BOUND_TIME = 2

BETTER_UNKNOWN = 0
BETTER_ALWAYS_LESS = 1
BETTER_ALWAYS_MORE = 2
BETTER_LESS_IF_IO_BOUND = 3 # but more if time bound
BETTER_MORE_IF_IO_BOUND = 4 # but less if time bound

WIDE_DATA_THRESHOLD = 6

def better_info(bounds, better):
    if better == BETTER_ALWAYS_LESS:
        less_is_better = True

    elif better == BETTER_ALWAYS_MORE:
        less_is_better = False

    elif better == BETTER_LESS_IF_IO_BOUND:
        if bounds == TEST_BOUND_IO:
            less_is_better = True
        else:
            less_is_better = False

    elif better == BETTER_MORE_IF_IO_BOUND:
        if bounds == TEST_BOUND_IO:
            less_is_better = False
        else:
            less_is_better = True

    else:
        return ('', '')

    if less_is_better:
        return ('&darr;', 'less is better')
    else:
        return ('&uarr;', 'more is better')

CONST_TIME_EXCUSE = " as this workload is time constrained"

def get_tracedir(resdir, workload, nfsopt):
    tracedir_root = '%s-%s-%s' % (TRACE_DIR_PREFIX, workload, nfsopt)
    tracedir_root = tracedir_root.replace('>', '')
    tracedir_root = tracedir_root.replace(' ', '_')
    tracedir_root = tracedir_root.replace('|', '_')
    tracedir_root = tracedir_root.replace('"', '_')
    tracedir_root = tracedir_root.replace("'", '_')

    i = 0
    while True:
        tracedir = '%s-%s' % (tracedir_root, i)

        if not os.path.exists(os.path.join(resdir, tracedir)):
            return tracedir

        i += 1

COLORS = [
  '#FEC44F', # Yellow
  '#D95F0E', # Orange
  '#476CDA', # Blue
  '#336600', # Green
  '#008B8B', # Turquoise
  '#303030', # Blackish
  '#FEE0B6',
  '#B2ABD2',
  '#8073AC',
  '#542788',
  '#2D004B',
  '#67001F',
  '#B2182B',
  '#D6604D',
  '#F4A582',
  '#FDDBC7',
  '#E0E0E0',
  '#BABABA',
  '#878787',
  '#4D4D4D',
  '#1A1A1A',
]

HATCHES = ['/', '.', 'x', '|', 'o', '-', '+', '\\', 'O', '*', ]

def hatch_idx(i):
    assert i > 0
    return (i - 1) % len(HATCHES)

def get_hatch(i):
    if i > 0:
        return HATCHES[hatch_idx(i)]
    return ''

def get_hatch_html(i, title=''):
    hatch = get_hatch(i)
    if not hatch:
        hatch = '&nbsp;'
    return '<span class="hatch" title="%s">%s</span>' % (title, hatch,)

def color_idx(i):
    return i % len(COLORS)

try:
    import numpy as np
except:
    import_error("Error importing numpy - Make sure numpy is installed")

try:
    import matplotlib
except:
    import_error("Error importing matplotlib - Make sure matplotlib is installed")

def check_mpl_version():
    vers = matplotlib.__version__

    warning = False
    sv = vers.split('.')

    if int(sv[0]) < 1:
        warning = True
    elif int(sv[0]) == 1 and int(sv[1]) < 1:
        warning = True

    if warning:
        warn("matplotlib version %s < 1.1 - some graph features might not work!" % vers)

try:
    # Don't require $DISPLAY to be set!
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
    import matplotlib.font_manager as fm
except:
    import_error("Error importing matplotlib submodules - this is probably an incompatible version of matplotlib")

try:
    from mako.template import Template
except:
    import_error("Error importing mako - Make sure mako is installed")

PLUSMINUS = '&#177;'

NO_DATA_HTML='<span class="no_data">no data</span>'
COMPARISON_ZERO_HTML='<span class="zero_data">none</span>'

def pluralize(x, pluralstr='s'):
    if x != 1:
        return pluralstr
    return ''

