"""
Copyright 2012 NetApp, Inc. All Rights Reserved,
contribution by Weston Andros Adamson <dros@netapp.com>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
"""

#!/usr/bin/env python

import multiprocessing
import cPickle

import os, sys, time
import traceback
from math import sqrt, pow
import hashlib

from collection import *
from config import *

def obj_tree_same(a, b):
    assert type(a) == type(b)
    if isinstance(a, dict):
        assert len(a) == len(b)
        for k in a.keys():
            obj_tree_same(a[k], b[k])
    elif isinstance(a, (list, tuple)):
        assert len(a) == len(b)
        for i in range(len(a)):
            obj_tree_same(a[i], b[i])
    else:
        assert a == b, "%r != %r" % (a, b)

def _small_keys(keys):
    """ attempt to transform a list of keys so that any substrings common
        to the start or end of all keys is removed """
    split_keys = []

    for k in keys:
        split_keys.append(k.split('_'))

    minwords = min([ len(x) for x in split_keys ])

    remove_idx_list = []

    for i in range(minwords):
        thisword_same = True

        for skey in split_keys[1:]:
            if split_keys[0][i] != skey[i]:
                thisword_same = False
                break

        if thisword_same:
            remove_idx_list.append(i)

    # remove words that are the same all across
    for idx in remove_idx_list[::-1]:
        for i in range(len(split_keys)):
            split_keys[i].pop(idx)

    for i in range(len(split_keys)):
        split_keys[i] = '_'.join(split_keys[i])

    return split_keys

class Stack(list):
    pass

class GraphFactory:
    def __init__(self, imagedir, serial_gen=False, debug_nographs=False):
        self.pool = None
        self.count = 0
        self.total = 0
        self.prune = 0
        self.prune_hash = 0
        self.prune_other = 0
        self.mismatch = 0
        self.serial_gen = serial_gen
        self.debug_nographs = debug_nographs
        self.imagedir = imagedir
        self.num_proc = max(multiprocessing.cpu_count() - 2, 2)

        try:
            os.mkdir(self.imagedir)
        except OSError, e:
            assert e.errno == os.errno.EEXIST

        self._data_map = {}
        self._data_rmap = {}

    def _get_mapping(self, data_str):
        assert not self._data_map.has_key(data_str), \
            "Graph data not unique %s" % data_str

        m = hashlib.md5()
        m.update(data_str)
        hexdigest = m.hexdigest()

        assert not self._data_rmap.has_key(hexdigest), \
            "Graph hash not unique"

        self._data_map[data_str] = hexdigest
        self._data_rmap[hexdigest] = data_str
        return hexdigest

    def pool_done(self, res):
        self.count -= 1


    def make_data_obj(self, data_tuple, args, kwargs):
        return {
            'config_mtime': os.stat(CONFIG_MODULE_FILE)[stat.ST_MTIME],
            'graph_mod_mtime': os.stat(__file__)[stat.ST_MTIME],
            'data_tuple': data_tuple,
            'args': args,
            'kwargs': kwargs,
        }

    def make_graph(self, graphtype, data_tuple, *args, **kwargs):
        data_obj = self.make_data_obj(data_tuple, args, kwargs)
        data_str = repr(data_obj)

        if graphtype == 'bar':
            graphfunc = make_bargraph_cb
        else:
            raise RuntimeError('Unhandled graphtype: %r' % graphtype)

        hexdigest = self._get_mapping(data_str)
        imgfile = os.path.join(self.imagedir,
                               '%s_%s.png' % (graphtype, hexdigest))
        datafile = os.path.join(self.imagedir,
                                '%s_%s.data' % (graphtype, hexdigest))

        imgtag = '<img src="./images/%s">' % os.path.split(imgfile)[-1]

        # see if the same graph already exists
        if os.access(imgfile, os.R_OK):
            try:
                f = file(datafile, 'r')
                other_data = cPickle.load(f)
                del f
            except:
                other_data = ''

            if data_obj == other_data:
                # graph is already made!
                return imgtag

            if self.debug_nographs:
                obj_tree_same(data_obj, other_data)

            try:
                os.unlink(imgfile)
                self.mismatch += 1
            except:
                pass
            try:
                os.unlink(datafile)
                self.mismatch += 1
            except:
                pass
        else:
            assert not self.debug_nographs, 'debug_nographs: file not found'

        args = list(args)
        args.insert(0, imgfile)
        args.insert(1, datafile)
        args.insert(2, data_obj)
        args = tuple(args)

        if self.serial_gen:
            graphfunc(*args)
        else:
            if self.pool == None:
                self.pool = multiprocessing.Pool(processes=self.num_proc)

            self.pool.apply_async(graphfunc, args, {}, self.pool_done)
            self.count += 1
            self.total += 1

        return imgtag

    def prune_graphs(self):
        for dentry in os.listdir(self.imagedir):
            hexdigest = None
            is_png = False
            if dentry.endswith('.png'):
                hexdigest = dentry[:-1 * len('.png')].split('_', 1)[1]
                is_png = True

            elif dentry.endswith('.data'):
                hexdigest = dentry[:-1 * len('.data')].split('_', 1)[1]

            if not hexdigest:
                os.unlink(os.path.join(self.imagedir, dentry))
                self.prune_other += 1
                continue

            if not self._data_rmap.has_key(hexdigest):
                os.unlink(os.path.join(self.imagedir, dentry))
                if is_png:
                    self.prune += 1
                else:
                    self.prune_hash += 1

    def wait_for_graphs(self):
        self.prune_graphs()

        if self.pool:
            self.pool.close()

            last_count = None
            while self.count > 0:
                if last_count != None and last_count == self.count:
                    print "Probable error with graph generation"

                last_count = self.count

                sys.stdout.write("\rGenerating graphs - (%u to go)......" %
                                 (self.count,))
                sys.stdout.flush()
                time.sleep(1)

            self.pool.join()

        inform('\rGraph Summary:                                  ')
        print '  %u images generated' % self.total
        print '  %u cached' % (len(self._data_map) - self.total)
        print '  %u data mismatch' % self.mismatch
        print '  %u images pruned' % self.prune
        print '  %u hash files pruned' % self.prune_hash
        print '  %u other files pruned' % self.prune_other

def make_bargraph_cb(imgfile, data_file, data_obj, units, vals, groups, keys,
                     no_ylabel, tall_cell):

    large_graph = (len(keys) * len(groups)) > WIDE_DATA_THRESHOLD

    # skip over any groups that dont exist in vals
    groups = [ g for g in groups if vals.has_key(g) ]

    matplotlib.rc('ytick', labelsize=8)
    matplotlib.rc('xtick', labelsize=8)

    fig = plt.figure(1)
    plt.clf()

    # small graphs
    graph_height = max(0.5 * len(groups), 0.7 * len(keys))
    graph_width = 2.5

    # if more than N bars, it's a large graph
    if large_graph:
        graph_height = 2.0
        graph_width = 7.0

    plt.gcf().set_size_inches(graph_width, graph_height)

    # ajust zoom
    pos_l = 0.01
    pos_r = 0.99
    pos_t = 0.99
    pos_b = 0.01

    if not no_ylabel:
        # make room for labels on the left
        pos_l = 0.25
        # also, tick values are "taller" than top of graph
        pos_t = 0.9

        if large_graph:
            pos_l = 0.1
            pos_t = 0.92

    if len(keys) > 1:
        # make room for key labels
        pos_b = 0.15
        if large_graph:
            pos_b = 0.075

    fig.subplots_adjust(left=pos_l, right=pos_r, top=pos_t, bottom=pos_b)

    ax = plt.subplot(111)

    # width of bars within a group
    width = 0.85 / len(groups)
    # space between bars, two extra - for prespace
    groupspace = 0.15 / (len(groups) + 2)

    # before each grouping of bars (per key)
    prespace = groupspace * 1.5
    space_multiplier = (groupspace + width)


    def _fmt_data(x):
        if isinstance(x, (list, tuple)):
            x = [ float(y) for y in x ]
            return np.mean(x), np.std(x)

        elif isinstance(x, (float, int, long)):
            return x, 0.0

        elif x == None:
            # when graphing, no data can just be zero
            return 0.0, 0.0

        raise ValueError('Unexpected data type for key %s: %r' %
                        (key, val))

    for i, col in enumerate(groups):
        # both map key -> hidx -> list of values
        valmap = {}
        errmap = {}

        max_hatch_index = -1
        for key in keys:
            if not key in valmap:
                valmap[key] = {}
            if not key in errmap:
                errmap[key] = {}

            val = vals[col].get(key, None)
            hidx = 0 # default hatch
            if isinstance(val, Stack):
                for s in val:
                    x_v, x_s = _fmt_data(s['value'])
                    hidx = s['hidx']

                    assert not valmap[key].has_key(hidx)
                    assert not errmap[key].has_key(hidx)
                    valmap[key][hidx] = x_v
                    errmap[key][hidx] = x_s
            else:
                x_v, x_s = _fmt_data(val)
                valmap[key][hidx] = x_v
                errmap[key][hidx] = x_s

            max_hatch_index = max(max_hatch_index, hidx)

        assert max_hatch_index >= 0
        assert len(valmap) == len(errmap)

        # get offsets for groups in array form - ie [0.0, 1.0, 2.0] ...
        ind = np.arange(len(keys))

        adj = prespace + (space_multiplier * float(i))
        # add to array to account for bars and spacing
        pos = ind + adj

        bottom = None

        for hidx in range(max_hatch_index + 1):
            this_data = []
            this_yerr = []
            for k in keys:
                this_data.append(valmap[k].get(hidx, 0.0))
                this_yerr.append(errmap[k].get(hidx, 0.0))

            # old versions of matplotlib dont support error_kw
            try:
                plt.bar(pos, this_data, width,
                        yerr=this_yerr,
                        bottom=bottom,
                        color=COLORS[color_idx(i)],
                        edgecolor='#000000',
                        hatch=get_hatch(hidx),
                        error_kw=dict(elinewidth=GRAPH_ERRORBAR_WIDTH,
                                      ecolor=GRAPH_ERRORBAR_COLOR))

            except AttributeError:
                plt.bar(pos, this_data, width,
                        yerr=this_yerr,
                        bottom=bottom,
                        color=COLORS[color_idx(i)],
                        edgecolor='#000000',
                        hatch=get_hatch(hidx))


            if not bottom:
                bottom = this_data
            else:
                assert len(bottom) == len(this_data)
                for bx in range(len(bottom)):
                    bottom[bx] += this_data[bx]

    # add some
    if not no_ylabel and units != None:
        plt.ylabel(units, size=8)
    else:
        plt.yticks([])

    # only write key names on x axis if there is more than one key
    if len(keys) > 1:
        plt.xticks(ind + (width * (float(len(groups))/2.0)),
                   [ x for x in _small_keys(keys) ], size=8)
    else:
        plt.xticks(ind, [''] * len(ind))

    # DONT ZOOM!
    plt.xlim(0, ind[-1] + 1.0)

    plt.savefig(imgfile, transparent=True)
    plt.close(1)
    f = file(data_file, 'w+')
    cPickle.dump(data_obj, f)
