"""
Copyright 2012 NetApp, Inc. All Rights Reserved,
contribution by Weston Andros Adamson <dros@netapp.com>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
"""

import getopt
import re
import os, posix, stat, sys

from config import *

# options class for parsing command line
OTYPE_BOOL=1
OTYPE_ARG=2
OTYPE_LIST=3
OTYPE_HELP=4

# no arguments expected
OMODES_ARG_NONE=0
# modes that need <server:path> and may have <workloads...>
OMODES_ARG_SERVERPATH=1
# modes that need <server:path> and one <workload>
OMODES_ARG_SERVERPATH_AND_ONE_WORKLOAD=2
# modes that need <server:path> and may have <workloads...>
OMODES_ARG_WORKLOAD=3

OMODES = {
 'all':    OMODES_ARG_SERVERPATH,

# included in all
 'fetch':  OMODES_ARG_WORKLOAD,
 'trace':  OMODES_ARG_SERVERPATH,
 'report': OMODES_ARG_NONE,

# not included in all
 'workloads': OMODES_ARG_NONE,
 'list':      OMODES_ARG_NONE,
 'loadgen':   OMODES_ARG_SERVERPATH_AND_ONE_WORKLOAD,
}
OMODE_DEFAULT='all'

class Options:

    # opts
    mode = None
    resultdir = RESULTS_DIR
    num_runs = 1
    options = []
    serial_graph_gen = False
    always_options = None
    randomize_traces = False
    tags = []

    serverpath = None
    workloads_requested = []

    _options_def = [
        ('r',  'resultdir',     OTYPE_ARG,  'resultdir',
         ("the directory used to save results",),
         "dir"),

        ('o',  'options',     OTYPE_LIST,  'options',
         ("mount options to iterate through",
          "this option may be used multiple times",
          "space separated arg also defines multiple options",
          "each mount option must have a version specified",),
         "mount.nfs options"),

        ('a', 'always-options', OTYPE_ARG, 'always_options',
         ("options added to every trace",
          "this option may be used multiple times",),
         'mount.nfs options'),

        ('t', 'tag', OTYPE_LIST, 'tags',
         ("tag all traces with 'tags'",
          "this option may be used multiple times",
          "tags can also be comma separated",),
         'tags'),

        ('n',  'num-runs',      OTYPE_ARG,  'num_runs',
         ("number of runs for each trace of",
          "<options> X <tags> X <workloads>",),
         "num runs"),

        (None, 'serial-graphs', OTYPE_BOOL, 'serial_graph_gen',
         ("generate graphs inline while generating reports.",
          "useful for debugging graphing issues",),
         None),

        (None, 'rand', OTYPE_BOOL, 'randomize_traces',
         ("randomize the order of traces",),
         None),

        ('h', 'help', OTYPE_HELP, None,
         ("show the help message",),
         None),
    ]

    def _getopt_short(self):
        ret = ''
        for oshort, olong, otype, oname, ohelp, odesc in self._options_def:
            if oshort:
                assert len(oshort) == 1, 'multi character short option!'
                if otype in (OTYPE_ARG, OTYPE_LIST):
                    ret += oshort + ':'
                else:
                    ret += oshort
        return ret

    def _getopt_long(self):
        ret = []
        for oshort, olong, otype, oname, ohelp, odesc in self._options_def:
            if olong:
                if otype in (OTYPE_ARG, OTYPE_LIST):
                    ret.append(olong + '=')
                else:
                    ret.append(olong)
        return ret

    def parse(self):
        shortstr = self._getopt_short()
        longlist = self._getopt_long()

        try:
            opts, args = getopt.getopt(sys.argv[1:], shortstr, longlist)

        except getopt.GetoptError, err:
            self.usage(str(err))

        # parse options
        for o, a in opts:
            found = False
            for oshort, olong, otype, oname, ohelp, odesc in self._options_def:
                if (oshort and o == '-' + oshort) or \
                   (olong and o == '--' + olong):
                    if otype == OTYPE_BOOL:
                        setattr(self, oname, True)
                    elif otype == OTYPE_ARG:
                        setattr(self, oname, a)
                    elif otype == OTYPE_LIST:
                        getattr(self, oname).append(a)
                    elif otype == OTYPE_HELP:
                        self.usage()
                    else:
                        raise ValueError('Invalid OTYPE: %u' % (otype,))

                    found = True
                    break
            if not found:
                self.error('Invalid option: %s' % (o,))

        # parse and validate args

        # parse mode
        if len(args) >= 1 and args[0] in OMODES:
            self.mode = args[0]
            args = args[1:]
        else:
            self.mode = OMODE_DEFAULT

        mode_arg_type = OMODES[self.mode]

        if mode_arg_type == OMODES_ARG_SERVERPATH:
            # <server:path> [<workload_1> ... <workload_N>]
            if not len(args):
                self.error('missing <server:path> argument')

            if args[0].find(':') < 0:
                self.error("<server:path> argument expected, "
                           "but no ':' found: %r" % args[0])

            self.serverpath = args[0]
            self.workloads_requested = args[1:]
            args = []

        elif mode_arg_type == OMODES_ARG_SERVERPATH_AND_ONE_WORKLOAD:
            # <server:path> <workload>
            if not len(args):
                self.error('missing <server:path> argument')

            if args[0].find(':') < 0:
                self.error("<server:path> argument expected, "
                           "but no ':' found: %r" % args[0])

            self.serverpath = args[0]
            args = args[1:]

            if not len(args):
                self.error("expecting workload argument after <server:path>")
            if len(args) > 1:
                self.error("expecting only one workload argument after"
                           "<server:path>")
            self.workloads_requested = args
            args = []

        elif mode_arg_type == OMODES_ARG_WORKLOAD:
            self.workloads_requested = args
            args = []

        elif mode_arg_type == OMODES_ARG_NONE:
            if len(args):
                self.error("unexpected arguments: %s" % (' '.join(args),))

        else:
            raise ValueError("unhandled mode_arg_type %r" % (mode_arg_type,))

        # normalize
        if not self.options and mode_arg_type == OMODES_ARG_SERVERPATH:
            inform('No options specified. '
                        'Using default: -o v3 -o v4 -o v4.1')
            self.options = ['v3', 'v4', 'v4.1']

        elif not self.options and \
            mode_arg_type == OMODES_ARG_SERVERPATH_AND_ONE_WORKLOAD:
            inform('No options specified. '
                        'Using %s default: -o v4' % (self.mode,))
            self.options = ['v4',]

        elif self.options and not mode_arg_type in \
            (OMODES_ARG_SERVERPATH, OMODES_ARG_SERVERPATH_AND_ONE_WORKLOAD):
            self.error('options are not allowed for mode %s' % (self.mode,))

        mountopts = []
        for x in self.options:
            mountopts.extend(re.split('[ |]', x))

        if self.always_options:
            mountopts = [ x + ',' + self.always_options for x in mountopts ]

        errors = []
        self.mountopts = []
        for x in mountopts:
            try:
                norm = nfsopt_normalize(x, require_version=True)
                self.mountopts.append(norm)
            except Exception, e:
                errors.append(str(e))

        if errors:
            self.error('\n'.join(errors))

        self.num_runs = int(self.num_runs)

        self.server = None
        self.path = None

        if self.serverpath:
            self.server, self.path = self.serverpath.split(':', 1)

        err = validate_tags(self.tags)
        if err:
            self.error('\n'.join(err))

        if mode_arg_type == OMODES_ARG_SERVERPATH_AND_ONE_WORKLOAD:
            if not len(self.mountopts) == 1:
                self.error("mode %s expects only one option", (self.mode,))

    def _option_help(self):
        lines = []
        for oshort, olong, otype, oname, ohelp, odesc in self._options_def:
            if not odesc:
                odesc = ''
            optstrs = []
            if oshort:
                ods = ''
                if odesc:
                    ods = ' <%s>' % odesc
                optstrs.append('-' + oshort + ods)
            if olong:
                ods = ''
                if odesc:
                    ods = '="<%s>"' % odesc
                optstrs.append('--' + olong + ods)
            optstrs = ', '.join(optstrs)

            if oname:
                val = getattr(self, oname)
            else:
                val = None
            if val:
                defaultstr = 'default: %r' % (val,)
            else:
                defaultstr = ''

            # function to split ohelp to fit on 80 column screen
            def _fmthelp(chunks, offset):
                ret = []
                for x in chunks:
                    ret.append(x)

                rfmt = '\n' + (' ' * offset)
                return rfmt.join(ret)

            lines.append('%s' % (optstrs,))
            lines.append('%-15s%s' % ('', _fmthelp(ohelp, 17)))
            if defaultstr:
                lines.append('%-15s%s' % ('', defaultstr))
            lines.append('')

        return lines

    def error(self, msg=''):
        print >>sys.stderr, msg
        print >>sys.stderr, '\nrun "%s --help" for more info' % sys.argv[0]
        sys.stderr.flush()
        sys.exit(1)

    def usage(self, msg=''):
        print >>sys.stderr, \
            """
usage: %(script)s [mode] [options] 

Basic usage (no mode specified):

 %(script)s <server:path> [workloads...] 

  - this will fetch needed files, run traces, then generate reports -- the
    'fetch', 'trace' and 'report' stages described below.

Advanced usage (specify modes):

 %(script)s list

    - list the contents of the results directory

 %(script)s workloads

    - list available and unavailable workloads

 %(script)s loadgen <server:path> <workload>

    - run in loadgen mode which doesn't record any stats and loops over
      <workload> against <server:path>.  Only one option is allowed.

 %(script)s fetch [workloads...]

    - fetch all needed files for the specified workload(s).  If no workloads
      are specified, all workloads are fetched.  Fetched files are only
      downloaded once and are cached for future runs.

 %(script)s trace <server:path> [workloads...]

    - run traces against <server:path>.  The traces run will be:
      (options + always options + tags) X (workloads) X (num runs)
      This will only run traces that don't already exist in the results
      directory.

 %(script)s report

    - generate all reports available from the results directory.

        """.strip() % {'script': sys.argv[0]}

        print >>sys.stderr
        print >>sys.stderr, "Options:"
        print >>sys.stderr, '  %s' % '\n  '.join(self._option_help())

        if msg:
            print >>sys.stderr
            print >>sys.stderr, "Error: " + msg

        sys.exit(1)

