"""
Copyright 2012 NetApp, Inc. All Rights Reserved,
contribution by Weston Andros Adamson <dros@netapp.com>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
"""

from config import *

SELECTOR_ORDER=(
    'workload',
    'kernel',
    'nfsopt',
    'client',
    'server',
)

_valid_things = set(SELECTOR_ORDER)

_SELECTOR_NORM={
    'nfsopt': nfsopt_normalize,
}

def _norm(name, x):
    _norm_f = _SELECTOR_NORM.get(name, None)

    if not _norm_f:
        return x

    if isinstance(x, (list, tuple)):
        return [ _norm_f(y) for y in x ]
    return _norm_f(x)

_SELECTOR_FMT={
    'kernel':   get_kernel_display,
    'nfsopt':   get_nfsopt_display,
    'workload': get_workload_display,
}

_SELECTOR_FMT_LONG={
    'nfsopt':   get_nfsopt_display_long,
}

def _fmt(name, x, default=None, short=True):
    _fmt_f = _SELECTOR_FMT.get(name, default)
    if not short and _SELECTOR_FMT_LONG.has_key(name):
        _fmt_f = _SELECTOR_FMT_LONG[name]
    if not _fmt_f:
        return x
    return _fmt_f(x)

class Selector(object):
    """ This class is used to specify selection of the current query
    """
    def __init__(self, *args):
        things = []

        assert len(args) == len(SELECTOR_ORDER)

        for idx, name in enumerate(SELECTOR_ORDER):
            obj = args[idx]
            if not isinstance(obj, (list, tuple)):
                obj = [obj]
            obj = _norm(name, obj)

            setattr(self, name + 's', tuple(obj))

    def __str__(self):
        out = []

        for name in SELECTOR_ORDER:
            obj = getattr(self, name + 's')
            out.append('%s%s: %s' % (name, pluralize(len(obj)), ', '.join(obj)))

        return ', '.join(out)

    def __hash__(self):
        args = []

        for name in SELECTOR_ORDER:
            obj = getattr(self, name + 's')
            assert len(obj) == 1, \
                "Can't hash selector with %s length != 1" % (name,)
            args.append(obj)

        return hash(tuple(args))

    def __cmp__(self, other):
        for name in SELECTOR_ORDER:
            r = cmp(getattr(self, name + 's'), getattr(other, name + 's'))
            if r:
                return r
        return 0

    def __repr__(self):
        args = []

        for name in SELECTOR_ORDER:
            obj = getattr(self, name + 's')
            args.append(obj)

        return 'Selector(%s)' % (', '.join([repr(x) for x in args]),)

    def __getattr__(self, attr):
        if attr in SELECTOR_ORDER:
            obj = getattr(self, attr + 's')
            assert len(obj) == 1, "%s is not singular" % attr
            return obj[0]
        elif hasattr(self, attr):
            return super(Selector, self).__getattr__(attr)
        else:
            raise AttributeError, "invalid attribute: %r" % attr

    def html(self):
        out = []

        for name in SELECTOR_ORDER:
            obj = getattr(self, name + 's')
            out.append('%s%s: %s' % (name, pluralize(len(obj)), ', '.join(obj)))

        return '<br>'.join(out)

    def is_valid_key(self):
        for name in SELECTOR_ORDER:
            obj = getattr(self, name + 's')
            if len(obj) != 1:
                return False
        return True

    def _foreach_thing(self, thing):
        assert thing

        if isinstance(thing, (list, tuple)):
            thing = list(thing)
            more_things = thing[1:]
            thing = thing[0]
        else:
            assert thing in _valid_things
            more_things = []

        for x in getattr(self, thing + 's'):
            args = []

            for name in SELECTOR_ORDER:
                if name == thing:
                    obj = x
                else:
                    obj = getattr(self, name + 's')
                args.append(obj)

            sel = Selector(*args)

            if more_things:
                for y in sel._foreach_thing(more_things):
                    yield y
            else:
                yield sel

    def foreach(self, thing=None):
        if thing == None:
            thing = SELECTOR_ORDER

        for x in self._foreach_thing(thing):
            yield x

    def foreach_nfs_version(self):
        vers_to_nfsopt = {}
        versions = []

        # maintain order of nfsopts between versions
        for x in self.nfsopts:
            vers = nfsopt_version(x)
            if not vers in versions:
                versions.append(vers)
            if not vers_to_nfsopt.has_key(vers):
                vers_to_nfsopt[vers] = []
            vers_to_nfsopt[vers].append(x)

        for vers in versions:
            sel = Selector(self.workload, self.kernels, vers_to_nfsopt[vers],
                           self.clients, self.servers)
            yield sel

    def fmt(self, thing, short=True):
        assert thing in _valid_things
        x = getattr(self, thing + 's')
        return _fmt(thing, x, default = lambda x : ' '.join(x), short=short)

    def display_info(self, all_selector):
        display_info = []

        for name in SELECTOR_ORDER:
            obj = getattr(self, name + 's')
            all_obj = getattr(all_selector, name + 's')

            if obj != all_obj:
                pl = pluralize(len(obj))
                entry = ('%s%s' % (name, pl), str(_fmt(name, obj)))
                display_info.append(entry)

        return display_info

