"""
Copyright 2012 NetApp, Inc. All Rights Reserved,
contribution by Weston Andros Adamson <dros@netapp.com>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
"""

TAGTYPE_MOUNTOPT_VERS=1
TAGTYPE_DETECTED=2
TAGTYPE_USER=3

TAG_DELEG_ENABLED='deleg'
TAG_PNFS_ENABLED='pnfs'

TAG_DESC = {
    'v2':                    TAGTYPE_MOUNTOPT_VERS,
    'v3':                    TAGTYPE_MOUNTOPT_VERS,
    'v4.0':                  TAGTYPE_MOUNTOPT_VERS,
    'v4.1':                  TAGTYPE_MOUNTOPT_VERS,

    TAG_DELEG_ENABLED: TAGTYPE_DETECTED,
    TAG_PNFS_ENABLED:  TAGTYPE_DETECTED,
}

def validate_tags(tags):
    r = []
    for t in tags:
        if t in TAG_DESC:
            r.append('%s is a reserved tag' % (t))
    return r

def make_nfsopt(mountopt, tags):
    nfsopt = mountopt

    if tags:
        nfsopt += ',%s' % ','.join(tags)

    return nfsopt

def _nfsopt_sort(nfsopts):
    """ Sort nfsopts, returns sorted list of nfsopts
    """
    splitopts = [ nfsopt_split(x) for x in nfsopts ]

    mountopts_data = {}
    mountopts_order = []
    for m, d, u in splitopts:
        m = ','.join(m)

        if not mountopts_data.has_key(m):
            mountopts_data[m] = []
        mountopts_data[m].append((d,u))

        if not m in mountopts_order:
            mountopts_order.append(m)

    mountopts_order.sort()

    for k in mountopts_data.keys():
        mountopts_data[k].sort()

    ret = []
    for m in mountopts_order:
        for d, u in mountopts_data[m]:
            tags = []
            tags.extend(d)
            tags.extend(u)
            ret.append(make_nfsopt(m, tags))

    return ret

def nfsopt_normalize(nfsopts, old_syntax=False, require_version=False):
    """
        normalize nfsopts, return a string normalized nfsopts

        if old_syntax is True, versions will not use vX.Y format and will
        instead use vers=X,minorversion=Y

        if require_version is True, raise an exception if a version isn't found
    """
    if isinstance(nfsopts, (list, tuple)):
        ret = []
        for n in nfsopts:
            ret.append(_nfsopt_normalize(n, old_syntax, require_version))

        return _nfsopt_sort(ret)
    else:
        return _nfsopt_normalize(nfsopts, old_syntax, require_version)

def _nfsopt_normalize(nfsopt, old_syntax, require_version):
    otheropts = []
    major = None
    minor = None

    for n in nfsopt.split(','):
        if n.startswith('vers='):
            val = n.split('=', 1)[1]

            if val.find('.') >= 0:
                val = val.split('.', 1)
                major = int(val[0])
                minor = int(val[1])
            else:
                major = int(val)

        elif n.startswith('minorversion='):
            val = n.split('=', 1)[1]
            minor = int(val)

        elif n.startswith('v'):
            val = n[1:]

            if val and val[0].isdigit() and val.find('.') >= 0:
                val = val.split('.', 1)
                major = int(val[0])
                minor = int(val[1])

            elif val.isdigit():
                major = int(val)

            else:
                otheropts.append(n)

        else:
            otheropts.append(n)

    vers = None
    if major == 2:
        if old_syntax:
            vers = 'vers=2'
        else:
            vers = 'v2'

    elif major == 3:
        if old_syntax:
            vers = 'vers=3'
        else:
            vers = 'v3'

    elif major == 4 and minor in (None, 0):
        if old_syntax:
            vers = 'vers=4'
        else:
            vers = 'v4.0'

    elif major == 4 and minor == 1:
        if old_syntax:
            vers = 'vers=4,minorversion=1'
        else:
            vers = 'v4.1'

    if vers:
        new = []
        otheropts.sort()
        new.append(vers)
        new.extend(otheropts)
        nfsopt = ','.join(new)

    elif require_version:
        raise ValueError("Version not found in nfsopt '%s'" % nfsopt)

    else:
        otheropts.sort()
        nfsopt = ','.join(otheropts)

    return nfsopt

def nfsopt_split(nfsopt):
    split_tags = nfsopt.split(',')

    assert split_tags

    version_tags = []
    server_tags = []
    client_tags = []

    for tag in split_tags:
        tagtype = TAG_DESC.get(tag, TAGTYPE_USER)

        if tagtype == TAGTYPE_MOUNTOPT_VERS:
            version_tags.append(tag)
        elif tagtype == TAGTYPE_DETECTED:
            server_tags.append(tag)
        elif tagtype == TAGTYPE_USER:
            client_tags.append(tag)
        else:
            raise NotImplemented

    r = (tuple(version_tags), tuple(server_tags), tuple(client_tags))
    return r

def nfsopt_version(nfsopt):
    """ return normalized string form of the NFS protocol version from 
        nfsopt
    """
    vers_list = nfsopt_split(nfsopt)[0]
    assert len(vers_list) == 1
    return vers_list[0]

def nfsopt_version_other(nfsopt):
    """ return a tuple where the first index is the normalized NFS proto
        version and the second index is the other parts of nfsopt
    """
    vers_list, one, two = nfsopt_split(nfsopt)
    assert len(vers_list) == 1, "vers_list = %r" % (vers_list,)
    both = []
    both.extend(one)
    both.extend(two)
    return (vers_list[0], ','.join(both))

